
// JPicsPanel.java
// Andrew Davison, March 2011, ad@fivedots.psu.ac.th

/* Take a picture every DELAY ms, and show in the panel.
   A good DELAY value can be obtained by looking at the 
   average ms time for obtaining a snap, which is written at
   the bottom of the panel.

   For Windows XP, the average time is 10-20ms, so I've
   set the delay to 25ms to give the thread a bit of a rest
   after each snap while the screen is repainted.

   The snaps are taken by a thread, which is terminated when the
   isRunning boolean is set to true by closeDown(). closeDown()
   blocks until run() has called JMFCapture.close() and finished. 
   This avoids nasty OS effects after the application has 
   terminated, but it means that clicking on the close box
   'freezes' the application for 1-2 secs.
*/

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.awt.image.*;
import java.text.DecimalFormat;
import java.io.*;
import javax.imageio.*;


class JPicsPanel extends JPanel implements Runnable
{
  private static final Dimension PANEL_SIZE = new Dimension(200, 100);  
          // dimensions of panel initially;  later set to video's frame size
  private static final double SCALE_FACTOR = 0.5;  // for image

  private static final int DELAY = 25;  // ms  (could be 10-20ms)

  private static final String SAVE_FNM = System.getProperty("user.home") + "/desktop/jmfPic.jpg";
      // locate saved snap on the Windows desktop (NOT portable)

  private JFrame top;
  private BufferedImage image = null;
  private JMFCapture camera; 
  private volatile boolean isRunning;
  
  // used for the average ms snap time info
  private int imageCount = 0;
  private long totalTime = 0;
  private DecimalFormat df;
  private Font msgFont;


  public JPicsPanel(JFrame top)
  {
    this.top = top;
    setBackground(Color.white);
    setMinimumSize(PANEL_SIZE);
    setPreferredSize(PANEL_SIZE);

    df = new DecimalFormat("0.#");  // 1 dp
    msgFont = new Font("SansSerif", Font.BOLD, 18);

    new Thread(this).start();   // start updating the panel's image
  } // end of JPicsPanel()



  public void run()
  /* take a picture every DELAY ms */
  {
    camera = new JMFCapture();

    // update panel and window sizes to fit video's frame size
    Dimension frameSize = camera.getFrameSize();
    if (frameSize != null) {
      setMinimumSize(frameSize);
      setPreferredSize(frameSize);
      top.pack();   // resize and center JFrame
      top.setLocationRelativeTo(null);
    }

    long duration;
    BufferedImage im = null;
    isRunning = true;

    while (isRunning) {
	    long startTime = System.currentTimeMillis();
      // System.out.println("Start loading image " + (imageCount+1) + "...");
      // im = scaleImage( camera.getImage(), SCALE_FACTOR);  // take a snap
      im = camera.getImage();  // take a snap

      duration = System.currentTimeMillis() - startTime;

      if (im == null)
        System.out.println("Problem loading image " + (imageCount+1));
      else {
        image = im;   // only update image if im contains something
        imageCount++;
        totalTime += duration;
        // System.out.println("Image " + imageCount + " loaded in " +
        //                     duration + " ms");
        repaint();
      }

      if (duration < DELAY) {
        try {
          Thread.sleep(DELAY-duration);  // wait until DELAY time has passed
        } 
        catch (Exception ex) {}
      }
    }

    // saveSnap(image, SAVE_FNM);   // save last image
    camera.close();    // close down the camera
  }  // end of run()


  public void paintComponent(Graphics g)
  /* Draw the snap and add the average ms snap time at the 
     bottom of the panel. */
  { 
    super.paintComponent(g);

    int panelHeight = getHeight();
    g.setColor(Color.WHITE);    // white background
    g.fillRect(0, 0, getWidth(), panelHeight);

    // center the image
    int x = 0;
    int y = 0;
    if (image != null) {
      x = (int)(getWidth() - image.getWidth())/2;
      y = (int)(panelHeight - image.getHeight())/2;
    }
    g.drawImage(image, x, y, this);   // draw the snap

    // write statistics in bottom-left corner
	  g.setColor(Color.blue);
    g.setFont(msgFont);
    if (imageCount > 0) {
      double avgGrabTime = (double) totalTime / imageCount;
	  g.drawString("Pic " + imageCount + "  " +
                   df.format(avgGrabTime) + " ms", 
                   5, panelHeight-10);  // bottom left
    }
    else  // no image yet
	    g.drawString("Loading...", 5, panelHeight-10);
  } // end of paintComponent()


  public void closeDown()
  /* Terminate run() and wait for the camera to be closed.
     This stops the application from exiting until everything
     has finished. */
  { 
    isRunning = false;
    while (!camera.isClosed()) {
      try {
        Thread.sleep(DELAY);
      } 
      catch (Exception ex) {}
    }
  } // end of closeDown()



 // ----------------- support methods ------------------------------


  private BufferedImage scaleImage(BufferedImage im, double scale)
  // Make a scaled copy of im, assuming no alpha channel
  {
    int scaledWidth = (int)((double)im.getWidth() * scale);
    int scaledHeight = (int)((double)im.getHeight() * scale);

    BufferedImage copy = new BufferedImage(scaledWidth, scaledHeight, 
                                        BufferedImage.TYPE_INT_RGB);
    // create a graphics context
    Graphics2D g2d = copy.createGraphics();

    // resize a copy of the image 
    g2d.setRenderingHint(RenderingHints.KEY_INTERPOLATION, 
                       RenderingHints.VALUE_INTERPOLATION_BILINEAR);
    g2d.drawImage(im, 0, 0, scaledWidth, scaledHeight, null);

    g2d.dispose();
    return copy;
  }  // end of scaleImage()


  private void saveSnap(BufferedImage im, String fnm)
  // save image as JPG
  {
    System.out.println("Saving image to " + fnm);
    try {
      ImageIO.write(im, "jpg", new File(fnm));
    }
    catch(IOException e)
    {  System.out.println("Could not save image");  }
  }  // end of saveSnap()

} // end of JPicsPanel class

